"""Core functions to perform interpolation by an arbitrary interpolation kernel
(such as in `.linear` or `.pchip`) in one dimension, with a variety of functions 
to facilitate interpolating with one or two kernels, one or two dependent data
sets, and/or to datasets that are 1D or multi-dimensional.  

The functions in this module are not intended to be called directly (though they
can be).  Rather, they are used by factory functions that return a new function
with the interpolation kernel (the input `f`) enclosed, thereby accelerating the
code.  See `.tools/make_interpolator`.

This subpackage is useful for interpolation when the evaluation site is known,
so that only one or a few interpolations are needed.  However, when 
interpolation must be repeated many times, such as when solving a nonlinear
equation involving the interpolants, the `ppinterp` subpackage is preferred, as 
that subpackage pre-computes the piecewise polynomial coefficients once, and then
interpolation to a given evaluation site is fast.

Note: In theory, many of these functions could be collapsed down to a single 
function.  For example, `_interp_1` and `_interp_1_YZ` could be replaced by
a single function that accepts a tuple as its fourth `Y` parameter; then it 
works like `_interp_1_YZ` over each element of `Y`.  However, as of Numba v0.53,
that approach is considerably *slower* than the approach taken here.
"""

import numpy as np
import numba as nb


@nb.njit
def _interp_1(f, x, X, Y):
    """
    Apply a given kernel of interpolation once.

    Parameters
    ----------
    f : function

        The "kernel" of interpolation: A function that performs a single
        interpolation within a known interval.

        The parameters to `f` are
        x : float
        X : ndarray(float, 1d)
        Y : ndarray(float, 1d)
        i : int
        and the return value of `f` is
        y : float
        which is `Y` as a function of `X` interpolated to the value `x`. Here,
        the subinterval of `X` within which `x` falls is given to this
        function by `i`.  This function will only be called when the
        following is guaranteed:
        (a) `i == 1`  and  `X[0] <= x <= X[1]`, or
        (b) `2 <= i <= len(X) - 1`  and  X[i-1] < x <= X[i]`.

    x : float
        Evaluation site

    X : ndarray(float, 1d)
        The independent data.
        Must be monotonically increasing.
        NaN is treated as +inf, hence NaN's must go after any valid data.
        If X[0] is NaN, it is assumed that all elements of X are NaN.

    Y : ndarray(float, 1d)
        The dependent data, with the same length as `X`.

    Returns
    -------
    y : float
        The value of `Y` interpolated to `X` at `x`.
    """
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        return np.nan

    # i = searchsorted(X,x) is such that:
    #   i = 0                   if x <= X[0] or all(isnan(X))
    #   i = len(X)              if X[-1] < x or isnan(x)
    #   X[i-1] < x <= X[i]      otherwise
    #
    # Having guaranteed above that
    # x is not nan, and X[0] is not nan hence not all(isnan(X)),
    # and X[0] <= x <= X[-1],
    # then either
    # (a) i == 0  and x == X[0], or
    # (b) 1 <= i <= len(X)-1  and  X[i-1] < x <= X[i]
    #
    # Next, merge (a) and (b) cases so that
    #   1 <= i <= len(X) - 1
    # is guaranteed, and
    #   X[0 ] <= x <= X[1]  when  i == 1
    #   X[i-1] < x <= X[i]  when  i > 1
    i = max(1, np.searchsorted(X, x))

    # Interpolate within the given interval
    return f(x, X, Y, i)


@nb.njit
def _interp_1_fg(f, g, x, X, Y):
    """As _interp_1 but applies two interpolation kernels."""
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        return np.nan, np.nan
    i = max(1, np.searchsorted(X, x))
    return f(x, X, Y, i), g(x, X, Y, i)


@nb.njit
def _interp_1_YZ(f, x, X, Y, Z):
    """As _interp_1 but applies the interpolation kernel to two dependent data arrays."""
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        return np.nan, np.nan
    i = max(1, np.searchsorted(X, x))
    return f(x, X, Y, i), f(x, X, Z, i)


@nb.njit
def _interp_1_fg_YZ(f, g, x, X, Y, Z):
    """As _interp_1 but applies two interpolation kernels to two dependent data arrays.

    Parameters
    ----------
    f, g : function
        As in `_interp_1`

    x, X : see `_interp_1`

    Y, Z : ndarray(float, 1d)
        As in `_interp_1`

    Returns
    -------
    yf : float
        The value of `Y` interpolated using `f` to `X` at `x`.

    zf : float
        The value of `Z` interpolated using `f` to `X` at `x`.

    yg: float
        The value of `Y` interpolated using `g` to `X` at `x`.

    zg : float
        The value of `Z` interpolated using `g` to `X` at `x`.

    """
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        return np.nan, np.nan, np.nan, np.nan
    i = max(1, np.searchsorted(X, x))
    return f(x, X, Y, i), f(x, X, Z, i), g(x, X, Y, i), g(x, X, Z, i)


@nb.njit
def _interp_n(f, x, X, Y):
    """
    As _interp_1 but applies interpolation kernel many times.

    Parameters
    ----------
    f, x, X : see `_interp_1`

    Y: ndarray(float, nd)

        Dependent data.  The last dimension must be the same length as `X`.

    Returns
    -------
    y : ndarray

        The value `y[n]` is `Y[n]` (a 1D array) interpolated to `X` at `x`.
        The shape of `y` is the shape of `Y` less its last dimension.

    Note
    ----
    This function is faster than a `numba.guvectorize`'d version of `_interp_1`
    because `numpy.searchsorted` is only called once, here.
    """
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        y = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
    else:
        y = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        i = max(1, np.searchsorted(X, x))
        for n in np.ndindex(y.shape):
            y[n] = f(x, X, Y[n], i)

    return y


@nb.njit
def _interp_n_YZ(f, x, X, Y, Z):
    """As _interp_n but applies the interpolation kernel to two dependent data ndarrays.
    Assumes Y and Z are the same shape.
    """
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        y = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
        z = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
    else:
        y = np.empty(Y.shape[0:-1], dtype=np.f8)
        z = np.empty(Y.shape[0:-1], dtype=np.f8)
        i = max(1, np.searchsorted(X, x))
        for n in np.ndindex(y.shape):
            y[n] = f(x, X, Y[n], i)
            z[n] = f(x, X, Z[n], i)

    return y, z


@nb.njit
def _interp_n_fg(f, g, x, X, Y):
    """As _interp_n but applies two interpolation kernels."""
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        yf = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
        yg = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
    else:
        yf = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        yg = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        i = max(1, np.searchsorted(X, x))
        for n in np.ndindex(yf.shape):
            yf[n] = f(x, X, Y[n], i)
            yg[n] = g(x, X, Y[n], i)

    return yf, yg


@nb.njit
def _interp_n_fg_YZ(f, g, x, X, Y, Z):
    """As _interp_n but applies two interpolation kernels to two dependent data ndarrays."""
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        yf = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
        zf = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
        yg = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
        zg = np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)
    else:
        yf = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        zf = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        yg = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        zg = np.empty(Y.shape[0:-1], dtype=Y.dtype)
        i = max(1, np.searchsorted(X, x))
        for n in np.ndindex(yf.shape):
            yf[n] = f(x, X, Y[n], i)
            zf[n] = f(x, X, Z[n], i)
            yg[n] = g(x, X, Y[n], i)
            zg[n] = g(x, X, Z[n], i)

    return yf, zf, yg, zg



"""Kernels for linear interpolation"""
import numba as nb


@nb.njit
def _linterp(x, X, Y, i):
    """
    The "kernel" of linear interpolation.

    Parameters
    ----------
    x : float
        The evaluation site

    X : ndarray(float, 1d)
        The independent data.

    Y : ndarray(float, 1d)
        The dependent data.

    i : int
        The interval of `X` that contains `x`.  Specifically,
        (a) `i == 1`  and  `X[0] <= x <= X[1]`, or
        (b) `2 <= i <= len(X) - 1`  and  X[i-1] < x <= X[i]`.
        These facts about `i` are assumed true; they are not checked.
        (This function will not be called if `x < X[0]` or `X[-1] < x` or
        `x` is nan or `X` are all nan.)

    Returns
    -------
    y : float
        The value of `Y` linearly interpolated to `X` at `x`.

    """

    # dx = (x - X[i-1]) => 0 guaranteed (and == 0 only if x == X[0])
    return (Y[i] - Y[i - 1]) / (X[i] - X[i - 1]) * (x - X[i - 1]) + Y[i - 1]


@nb.njit
def _linterp1(x, X, Y, i):
    """
    The "kernel" of the 1st derivative of linear interpolation.

    Inputs and outputs analogous to `_linterp`.
    """

    return (Y[i] - Y[i - 1]) / (X[i] - X[i - 1])


# interp_u = make_u(_linterp)
# interp_1 = make_1(_linterp)
# interp_n = make_n(_linterp)
# interp_u_twice = make_u_twice(_linterp)
# interp_1_twice = make_1_twice(_linterp)
# interp_n_twice = make_n_twice(_linterp)


# @nb.njit
# def interp_1_d(x, X, Y, d):
#     if d == 0:
#         return _interp_1(x, X, Y, ker)
#     elif d == 1:
#         return _interp_1(x, X, Y, ker1)
#     else:
#         return 0.0


# @nb.guvectorize(
#     [(float64, float64[:], float64[:], int64, float64[:])],
#     "(),(n),(n),()->()",
# )
# def interp_u_d(x, X, Y, d, y):
#     y[0] = interp_1_d(x, X, Y, d)


# # %timeit -n 100 interp_1_n(500., Z, S, 0)
# # 280 µs ± 23.1 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
# @nb.njit
# def interp_n_d(x, X, Y, d):
#     if d == 0:
#         return _interp_n(x, X, Y, ker)
#     elif d == 1:
#         return _interp_n(x, X, Y, ker1)
#     else:
#         return np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype)


# @nb.njit
# def interp_1_d_twice(x, X, Y, Z, d):
#     if d == 0:
#         return _interp_1_twice(x, X, Y, ker)
#     elif d == 1:
#         return _interp_1_twice(x, X, Y, ker1)
#     else:
#         return 0.0, 0.0


# @nb.guvectorize(
#     [(float64, float64[:], float64[:], float64[:], int64, float64[:], float64[:])],
#     "(),(n),(n),(n)()->(),()",
# )
# def interp_u_d_twice(x, X, Y, d, y, z):
#     y[0], z[0] = interp_1_d_twice(x, X, Y, Y, d)


# @nb.njit
# def interp_n_d_twice(x, X, Y, Z, d):
#     if d == 0:
#         return _interp_n_twice(x, X, Y, Z, ker)
#     elif d == 1:
#         return _interp_n(x, X, Y, Z, ker1)
#     else:
#         return (
#             np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype),
#             np.full(Y.shape[0:-1], np.nan, dtype=Y.dtype),
#         )


# Trying to make one function with different specializations, that behaves
# differently depending upon its input types.
# This is numba's polymorphic dispatching:
# https://numba.pydata.org/numba-doc/latest/developer/dispatching.html
# achieved using numba's generated_jit
# https://numba.pydata.org/numba-doc/dev/user/generated-jit.html
# Unfortunately this does not work because the function signature of interp_d,
# once under `guvectorize`, is utterly different from the function signature
# for interp_n_d or interp_universal...  This error:
#   TypeError: generated implementation <numba._GUFunc 'interp_d'> should be
#   compatible with signature '(x, X, Y, d, y)', but has signature
#   '(*args, **kwargs)'
# @nb.generated_jit(nopython=True)
# def interp_universal(x, X, Y, d, y):
#     if isinstance(x, nb.types.Float) and isinstance(X, nb.types.Array):
#         return interp_n_d
#     else:
#         return interp_d


# @nb.guvectorize(
#     [
#         (float32, float32[:], float32[:], float32[:]),
#         (float64, float64[:], float64[:], float64[:]),
#     ],
#     "(),(n),(n)->()",
# )
# def interp(x, X, Y, y):
#     """
#     A 'universal function' for 1D interpolation along the last dimension.

#     Parameters
#     ----------
#     x : ndarray or float

#         Sites at which to interpolate the independent data.

#         If ndarray, must be the same size as `X` or `Y` less their final
#         dimension, with any dimension possibly a singleton: that is, must be
#         broadcastable to the size of `X` or `Y` less their final dimension.

#     X : ndarray

#         Independent data. Must be same dimensions as `Y`, with any
#         dimension possibly a singleton: that is, must be broadcastable
#         to the size of `Y`.

#         `X` must be monotonically increasing in its last dimension.
#         That is, `X[*i,:]` must be monotonically increasing for any
#         `i` tuple indexing all but the last dimension.  NaN's in `X`
#         are treated as +Inf; hence, NaN's must go after any valid data.
#         (If X[0] is NaN, it is assumed that all elements of X are NaN.)

#     Y : ndarray

#         Dependent data.  Must be same dimensions as `X`, with any
#         dimension possibly a singleton: that is, must be broadcastable
#         to the size of `X`.

#     Returns
#     -------
#     y : ndarray or float

#         Value of the interpolant (or its `d`'th derivative) at `x`.

#         If `x` is float and `X` and `Y` are both 1 dimensional, `y` is float.

#         Otherwise, `y` is ndarray with size matching the largest of the sizes
#         of `x`, the size of `X` less its last dimension, or the size of `Y`
#         less its last dimension.

#     Notes
#     -----
#     A binary search is performed to find the indices of `X` that `x`
#     lies between.  As such, nonsense results will arise if `X` is not
#     sorted along its last dimension.
#     """
#     y[0] = _interp(x, X, Y, ker)



"""
Core functions to evaluate piecewise polynomial interpolants (or their
derivatives) in one dimension.  Additional functions are provided that do
this for two such interpolants, to facilitate the interpolation of paired
variables like salinity and temperature. 

This subpackage is ideally suited when interpolation to many different
evaluation sites is needed, such as when solving a nonlinear equation
involving the interpolants, because it pre-computes the piecewise polynomial
coefficients once, and then interpolation to a given evaluation site is fast.
However, if the interpolant needs to be evaluated just once or a few times,
the subpackage `interp1d` may be preferred.
"""


"""
Adapted from 'Piecewise Polynomial Calculus' available on the MATLAB
File Exchange at
https://mathworks.com/matlabcentral/fileexchange/73114-piecewise-polynomial-calculus
"""

import numpy as np
import numba as nb


@nb.njit
def ppval_i(dx, Yppc, i, d=0):
    """
    Evaluate a single interpolant, knowing where the evaluation site lies.

    Provides `i` such that  `X[i] < x <= X[i+1]`
    """

    # Evaluate polynomial, using nested multiplication.
    # E.g. the cubic case is:
    # y = dx^3 * Yppc[i,0] + dx^2 * Yppc[i,1] + dx * Yppc[i,2] + Yppc[i,3]
    #   = dx * (dx * (dx * Yppc[i,0] + Yppc[i,1]) + Yppc[i,2]) + Yppc[i,3]
    y = 0.0
    if d == 0:
        for j in range(0, Yppc.shape[1]):
            y = y * dx + Yppc[i, j]
    else:
        # Evaluate polynomial derivative, using nested multiplication.
        # E.g. the second derivative of the cubic case is:
        # y = 6 * dx * Yppc[i,0] + 2 * Yppc[i,1]
        y = 0.0
        degree = Yppc.shape[1] - 1
        for o in range(0, degree - d + 1):  # o for order
            p = 1.0  # Build integer multiplying the coefficient
            for a in range(degree - o - d + 1, degree - o + 1):
                p *= a
            # p = np.prod(np.arange(degree - o - d + 1, degree - o + 1))  # slow!
            y = y * dx + Yppc[i, o] * p

    return y


@nb.njit
def ppval1(x, X, Yppc, d=0):
    """
    Evaluate a single piecewise polynomial (PP).

    Parameters
    ----------
    x : float
        Evaluation site
    X : ndarray, 1d
        Independent variable.
    Yppc : ndarray, 2d
        Piecewise Polynomial Coefficients.  First dimension must match `len(X)`.
    d : int, Default 0
        Number of derivatives to take.  If 0, simply evaluate the PP.

    Returns
    -------
    y : float
        The value of the PP (or its `d`'th derivative) at `X = x`.

    """
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        return np.nan

    # i = searchsorted(X,x) is such that:
    #   i = 0                   if x <= X[0] or all(isnan(X))
    #   i = len(X)              if X[-1] < x or isnan(x)
    #   X[i-1] < x <= X[i]      otherwise
    #
    # Having guaranteed above that
    # x is not nan, and X[0] is not nan hence not all(isnan(X)),
    # and X[0] <= x <= X[-1],
    # then either
    # (a) i == 0  and x == X[0], or
    # (b) 1 <= i <= len(X)-1  and  X[i-1] < x <= X[i]
    #
    # Next, merge (a) and (b) cases so that
    #   1 <= i <= len(X) - 1
    # is guaranteed, and
    #   X[0 ] <= x <= X[1]  when  i = 1
    #   X[i-1] < x <= X[i]  when  i > 1
    # Then subtract 1 so that
    #   X[0 ] <= x <= X[1]   when  i = 0
    #   X[i]  <  x <= X[i+1] when  i > 0
    i = max(0, np.searchsorted(X, x) - 1)

    dx = x - X[i]  # >= 0
    return ppval_i(dx, Yppc, i, d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:, :], nb.i8, nb.f8[:])],
    "(),(n),(n,m),()->()",
)
def ppval(x, X, Yppc, d, y):
    """
    Evaluate a given Piecewise Polynomial

    Parameters
    ----------
    x : ndarray

        Sites at which to evaluate the independent variable.

        Must be broadcastable to the shape of `X` or `Y` less their final
        dimension.

    X : ndarray

        Independent variable. Must be broadcastable to the shape of `Y`.

        `X` must be monotonically increasing in its last dimension.
        That is, `X[*i,:]` must be monotonically increasing for any
        `i` tuple indexing all but the last dimension.  NaN's in `X`
        are treated as +Inf.

    Y : ndarray

        Dependent variable.  Must be broadcastable to the shape of `X`.

    Yppc : ndarray

        Piecewise Polynomial Coefficients for the interpolant.
        The shape of `Yppc` less its final dimension must be broadcastable to
        the shape of `X` and/or `Y`.

    Returns
    -------
    y : ndarray

        Value of the piecewise polynomial interpolant (or its `d`'th
        derivative) at `x`.

    Notes
    -----
    A binary search is performed to find the indices of `X` that `x`
    lies between.  As such, nonsense results will arise if `X` is not
    sorted along its last dimension.
    """
    y[0] = ppval1(x, X, Yppc, d)


@nb.njit
def ppval1_two(x, X, Yppc, Zppc, d=0):
    """
    Evaluate two piecewise polynomials.

    As `ppval1` but a second input, `Zppc`, provides a second set of Piecewise
    Polynomial Coefficients.  Correspondingly, a second output, `z`, is returned.
    """
    if np.isnan(x) or x < X[0] or X[-1] < x or np.isnan(X[0]):
        return np.nan, np.nan
    i = max(0, np.searchsorted(X, x) - 1)
    dx = x - X[i]  # >= 0
    return ppval_i(dx, Yppc, i, d), ppval_i(dx, Zppc, i, d)


@nb.guvectorize(
    [(nb.f8, nb.f8[:], nb.f8[:, :], nb.f8[:, :], nb.i8, nb.f8[:], nb.f8[:])],
    "(),(n),(n,m),(n,m),()->(),()",
)
def ppval_two(x, X, Yppc, Zppc, d, y, z):
    """
    Evaluate two given Piecewise Polynomials

    The inputs and outputs are as for `ppval`, but another input, `Zppc`, gives
    the second piecewise polynomial, for which a second output `z` is given.

    Notes
    -----
    Calling `ppval_two` is faster than calling `ppval` twice, since the former
    calls `np.searchsorted` half as many times.
    """
    y[0], z[0] = ppval1_two(x, X, Yppc, Zppc, d)


@nb.njit
def diff_1d_samesize(x):
    """
    First difference in one dimension; appends nan so output same size as input.

    A simple function like `np.diff` that numba can work with.
    """

    d = np.empty(x.size, dtype=x.dtype)
    for i in range(d.size):
        d[i] = x[i + 1] - x[i]
    d[-1] = np.nan
    return d



"""
Functions for linear or PCHIP interpolation of one (or two) dependent variables
in terms of one independent variable, done serially over an arbitrary number
of such interpolation problems.  This is used to interpolate salinity and
temperature in terms of either pressure or depth in each water column. 
"""
from .tools import make_interpolator, make_kernel



import numpy as np
import numba as nb

from .ppinterp import diff_1d_samesize


def linear_coeffs(X, Y):
    """
    Piecewise Polynomial Coefficients for a linear interpolant

    Parameters
    ----------
    X : ndarray

        Independent variable. Must be same dimensions as `Y`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `Y`.

        `X` must be monotonically increasing in its last dimension.
        That is, `X[*i,:]` should be monotonically increasing for any
        `i` tuple indexing all but the last dimension.  NaN's in `X`
        are treated as +Inf.

    Y : ndarray

        Dependent variable.  Must be same dimensions as `X`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `X`.

    Returns
    -------
    Yppc : ndarray

        Coefficients for a linear interpolant of `Y` to `X`.

    Notes
    -----

    Evaluate the piecewise polynomial at `x` as

    >>> y = ppval(x, X, Yppc)

    """
    return _linear_coeffs(X, Y, np.arange(2))


@nb.njit
def linear_coeffs_1(X, Y):
    """
    Coefficients for a single linear interpolant

    Inputs and outputs are as for `linear_coeffs`, but `X` and `Y` both 1D
    arrays of length n, and so `Yppc` is a 2D array of size (n, 2).
    """
    C = np.empty((len(Y), 2), dtype=np.float64)
    C[:, 0] = diff_1d_samesize(Y) / diff_1d_samesize(X)
    C[:, 1] = Y
    return C


@nb.guvectorize([(nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:, :])], "(n),(n),(m)->(n,m)")
def _linear_coeffs(X, Y, two, C):
    C[:, :] = linear_coeffs_1(X, Y)



from .linear import linear_coeffs, linear_coeffs_1
from .pchip import pchip_coeffs, pchip_coeffs_1


def select_ppc(interpolant="linear", kind="u"):
    """Select function for building piecewise polynomial coefficients."""
    if kind not in ("1", "u"):
        raise ValueError(f"Expected `kind` in ('1', 'u'); got {kind}")

    if interpolant == "linear":
        if kind == "1":
            return linear_coeffs_1
        elif kind == "u":
            return linear_coeffs
    elif interpolant == "pchip":
        if kind == "1":
            return pchip_coeffs_1
        elif kind == "u":
            return pchip_coeffs
    else:
        raise ValueError(
            f"Expected `interpolant` in ('linear', 'pchip'); got {interpolant}"
        )



# %% Imports

import numpy as np
import xarray as xr

# Functions to make the Equation of State
from neutralocean.eos import make_eos, make_eos_s_t

# Functions to compute various approximately neutral surfaces
from neutralocean.surface import potential_surf, anomaly_surf, omega_surf

from neutralocean.grid.xgcm import build_grid, edgedata_to_maps
from neutralocean.ntp import ntp_ϵ_errors

# %% Load data

print(
    "To get started, download the ECCOv4r4 grid information at\n"
    "  https://archive.podaac.earthdata.nasa.gov/podaac-ops-cumulus-protected/ECCO_L4_GEOMETRY_LLC0090GRID_V4R4/GRID_GEOMETRY_ECCO_V4r4_native_llc0090.nc"
    "\nand one day of Salinity and Temperature 3D data at\n"
    "  https://archive.podaac.earthdata.nasa.gov/podaac-ops-cumulus-protected/ECCO_L4_TEMP_SALINITY_LLC0090GRID_DAILY_V4R4/OCEAN_TEMPERATURE_SALINITY_day_mean_2002-12-23_ECCO_V4r4_native_llc0090.nc"
    "\nYou will have to create an Earthdata Login first."
    "Then edit the below variable, `folder_ecc4`, to point to the directory"
    " where you saved these two files."
)

folder_ecco4 = "~/work/data/ECCOv4r4/"  # << EDIT AS NEEDED >>
file_grid = folder_ecco4 + "GRID_GEOMETRY_ECCO_V4r4_native_llc0090.nc"
date = "2002-12-23"
file_ST = (
    folder_ecco4
    + "OCEAN_TEMPERATURE_SALINITY_day_mean_"
    + date
    + "_ECCO_V4r4_native_llc0090.nc"
)

# Load horizontal grid information
ds = xr.open_dataset(file_grid)
XC, YC, dxC, dyC, dxG, dyG = (
    ds[x].load() for x in ("XC", "YC", "dxC", "dyC", "dxG", "dyG")
)
ds.close()

# Load hydrographic data
ds = xr.open_dataset(file_ST)
S, T = (ds[x].squeeze().load() for x in ("SALT", "THETA"))

# Create depth xarray.
Z = -ds.Z.load()  # Make Z > 0 and increasing down
Z.attrs.update({"positive": "down"})  # Update attrs to match.

n = len(ds.i)  # size of each horizontal dimension in each square tile
ds.close()

# Get order of non-vertical dimensions
dims = tuple(x for x in S.dims if x != "k")

# Reorder dimensions to ensure individual water columns are float64 and contiguous in memory
S, T = (x.transpose(*dims, "k").astype(np.float64, order="C") for x in (S, T))

# define the connectivity between faces for the ECCOv4 LLC grid:
# fmt: off
face_connections = {'tile': {
    0: {'X':  ((12, 'Y', False), (3, 'X', False)),
        'Y':  (None,             (1, 'Y', False))},
    1: {'X':  ((11, 'Y', False), (4, 'X', False)),
        'Y':  ((0, 'Y', False),  (2, 'Y', False))},
    2: {'X':  ((10, 'Y', False), (5, 'X', False)),
        'Y':  ((1, 'Y', False),  (6, 'X', False))},
    3: {'X':  ((0, 'X', False),  (9, 'Y', False)),
        'Y':  (None,             (4, 'Y', False))},
    4: {'X':  ((1, 'X', False),  (8, 'Y', False)),
        'Y':  ((3, 'Y', False),  (5, 'Y', False))},
    5: {'X':  ((2, 'X', False),  (7, 'Y', False)),
        'Y':  ((4, 'Y', False),  (6, 'Y', False))},
    6: {'X':  ((2, 'Y', False),  (7, 'X', False)),
        'Y':  ((5, 'Y', False),  (10, 'X', False))},
    7: {'X':  ((6, 'X', False),  (8, 'X', False)),
        'Y':  ((5, 'X', False),  (10, 'Y', False))},
    8: {'X':  ((7, 'X', False),  (9, 'X', False)),
        'Y':  ((4, 'X', False),  (11, 'Y', False))},
    9: {'X':  ((8, 'X', False),  None),
        'Y':  ((3, 'X', False),  (12, 'Y', False))},
    10:{'X': ((6, 'Y', False),  (11, 'X', False)),
        'Y': ((7, 'Y', False),  (2, 'X', False))},
    11:{'X': ((10, 'X', False), (12, 'X', False)),
        'Y': ((8, 'Y', False),  (1, 'X', False))},
    12:{'X': ((11, 'X', False), None),
        'Y': ((9, 'Y', False),  (0, 'X', False))}}}
# fmt: on

# Number of Faces -- len(faces_connections only value)
nf = len(next(iter(face_connections.values())))

# Build list of adjacent water columns and distances between those water column pairs
xsh = ysh = "left"
grid = build_grid(n, face_connections, dims, xsh, ysh, dxC, dyC, dxG, dyG)


# Make Boussinesq version of the Jackett and McDougall (1995) equation of state
#  and its partial derivatives.
# TODO: double check that is what ECCOv4r4 used
grav, rho_c = 9.81, 1027.5
eos = make_eos("jmd95", grav, rho_c)
eos_s_t = make_eos_s_t("jmd95", grav, rho_c)

# Select pinning cast, picking the cast closest to (x0,y0)
x0, y0 = (-172, -4)  # longitude, latitude -- Pacific equatorial ocean
pin_cast = np.unravel_index(
    ((XC.values - x0) ** 2 + (YC.values - y0) ** 2).argmin(), XC.shape
)
z0 = 1500.0  # pinning depth

# %% Build approximately neutral surfaces!

# Build potential density surface, with given reference pressure (actually depth,
# for Boussinesq) and given isovalue.  No diagnostics requested, so info about
# the grid is not needed (no `edges` and `geoemtry` provided).
s, t, z, _ = potential_surf(
    S, T, Z, eos=eos, vert_dim="k", ref=0.0, isoval=1027.5, diags=False
)

# Build in-situ density anomaly surface with given reference salinity and
# potential temperature values and an isovalue of 0, which means the surface
# will intersect any point where the local (S,T) equals the reference values.
# Also return diagnostics.
s0, t0 = 34.5, 4.0
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="k",
    ref=(s0, t0),
    isoval=0.0,
)

# Build an omega surface that intersects the reference cast `pin_cast` at the
# reference depth `pin_p`, initialized from a potential density surface that
# also intersects this `pin_cast` at `pin_p` and uses a local reference
# depth, namely `pin_p`.
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid=grid,
    vert_dim="k",
    pin_cast=pin_cast,
    pin_p=z0,
    eos=(eos, eos_s_t),
    interp="pchip",
    ITER_MAX=10,
    ITER_START_WETTING=1,
)

# Calculate ϵ neutrality errors on the latest surface, between all pairs of adjacent water columns
ϵ = ntp_ϵ_errors(s, t, z, grid, eos_s_t)

# Convert the 1D array of ϵ values into two maps of ϵ neutrality errors, one
# for the errors in each of the two lateral ('i' and 'j') dimensions.
ϵi, ϵj = edgedata_to_maps(ϵ, n, face_connections, dims, xsh, ysh)



import functools as ft
import numba as nb
from numba import f4, f8

from .interp1d import _interp_1, _interp_1_YZ, _interp_1_fg, _interp_1_fg_YZ
from .interp1d import _interp_n, _interp_n_YZ, _interp_n_fg, _interp_n_fg_YZ
from .linear import _linterp, _linterp1
from .pchip import _pchip, _pchip1, _pchip2, _pchip3


@ft.lru_cache
def make_interpolator(interpolant="linear", deriv=0, kind="u", two=False):
    """Factory function to build various interpolating functions.

    Parameters
    ----------
    interpolant : str, Default "linear"

        If "linear", build a function for linear interpolation.
        If "pchip", build a function for Piecewise Cubic Hermite Interpolating
            Polynomial interpolation.

    deriv : int or tuple of int, Default 0

        Build a function that returns the `deriv` derivative of the
        `interpolant`.
        Note that `deriv = 0` simply builds a function that interpolates
        according to `interpolant`.
        If a tuple, gives a function that returns multiple values, one for each
        of the derivatives in this tuple.
        Currently, tuples of length 3 or more are not supported.

    kind : str, Default "u"

        If "1", a 'numba.njit'ed function is returned that does 1 interpolation.
        The function's inputs are (see `_interp_1`)
            - x : float
            - X : ndarray(float, 1d)
            - Y : ndarray(float, 1d)
            - Z : ndarray(float, 1d) -- only if `two` is True.

        If "n", a 'numba.njit'ed function is returned that does many
        interpolations of dependent data to a single evaluation site on a single
        independent data array.  The function's inputs are (see `_interp_n`)
            - x : float
            - X : ndarray(float, 1d)
            - Y : ndarray(float, nd)
            - Z : ndarray(float, nd) -- only if `two` is True; must have Z.shape == Y.shape

        If "u", a universal function is returned, whose inputs can be
        multidimensional numpy arrays, so long as they are appropriately
        broadcastable.  The function's inputs are
            - x : ndarray(float, (n-1)d)
            - X : ndarray(float, nd)
            - Y : ndarray(float, nd)
            - Z : ndarray(float, nd) -- only if `two` is True.

        If all dimensions are full, then the dimensions of `x`  match those of
        `X` less its last dimension, and all of `X`, `Y`, `Z` have the same
        dimension.
        However, these inputs dimensions can be a subset of their "full"
        dimensions above.  E.g. `X` can be a 1D array of length equal to the
        size of the last dimension of `Y`.
        This is achieved via `numba.guvectorize`ing the relevant `_interp_1_*`
        function.
        Most users will want this "u" option, as it can do everything the "1" or
        "n" options can do.  The exception is when a `@numba.njit` function is
        explicitly needed, such as when this is called inside another
        `@numba.njit` function.

    two : bool, Default False

        If True, returns an interpolating function that takes four inputs, the
        last two of which are two dependent data arrays, `Y` and `Z`.
        If False, just one dependent data array, `Y`, is input to the returned
        function.

    Returns
    -------
    f : function

        An interpolating function, taking three inputs,
            - `x`, the evaluation site,
            - `X`, the independent data, and
            - `Y`, the dependent data
        or a fourth input when `two` is True,
            - `Z`, another set of dependent data
        This function's output is determined by the inputs `deriv` and `two`.
        For example, with `deriv` = 0 and `two` = False, the output is
            - `y`, the interpolant of `Y` as a function of `X` evaluated at `x`.
        If `deriv` = 1 and `two` = True, then the outputs are
            - `y`, the 1st derivative for `Y` as a function of `X` evaluated at `x`,
            - `z`, the 1st derivative for `Z` as a function of `X` evaluated at `x`.
        If `deriv` = (0,1) and `two` = True, then the function's output is
            - `y0`, the interpolant of `Y` as a function of `X` evaluated at `x`,
            - `z0`, the interpolant of `Z` as a function of `X` evaluated at `x`,
            - `y1`, the 1st derivative for `Y` as a function of `X` evaluated at `x`,
            - `z1`, the 1st derivative for `Z` as a function of `X` evaluated at `x`.

    Examples
    --------
    >>> # Make fake pressure, salinity and temperature data
    >>> P = np.linspace(0, 4500, 10)
    >>> S = np.reshape([0, 1], (-1, 1)) + np.linspace(33.0, 35.0, 10).reshape((1, -1))
    >>> T = np.reshape([0, 1], (-1, 1)) + np.linspace(29.0, -2.0, 10).reshape((1, -1))
    >>> # Build a universal linear interpolator for two variables
    >>> interp = make_interpolator("linear", 0, "u", True)
    >>> p = 400.0  # interpolate to an isobaric surface (constant P)
    >>> s, t = interp(p, P, S, T)  # interpolate S and T to P = p.
    >>> p = [400.0, 450.0] # interpolate to a surface with varying pressure
    >>> s, t = interp(p, P, S, T)
    """

    ker0 = make_kernel(interpolant, deriv)

    if isinstance(ker0, tuple):
        lenker = len(ker0)
        if lenker == 2:
            ker1 = ker0[1]
        ker0 = ker0[0]  # Need to dereference these for numba
    else:
        lenker = 1

    if kind == "1":
        if lenker == 1:
            if not two:

                @nb.njit
                def fcn(x, X, Y):
                    return _interp_1(ker0, x, X, Y)

            else:

                @nb.njit
                def fcn(x, X, Y, Z):
                    return _interp_1_YZ(ker0, x, X, Y, Z)

        elif lenker == 2:
            if not two:

                @nb.njit
                def fcn(x, X, Y):
                    return _interp_1_fg(ker0, ker1, x, X, Y)

            else:

                @nb.njit
                def fcn(x, X, Y, Z):
                    return _interp_1_fg_YZ(ker0, ker1, x, X, Y, Z)

    elif kind == "n":

        if lenker == 1:
            if not two:

                @nb.njit
                def fcn(x, X, Y):
                    return _interp_n(ker0, x, X, Y)

            else:

                @nb.njit
                def fcn(x, X, Y, Z):
                    return _interp_n_YZ(ker0, x, X, Y, Z)

        elif lenker == 2:
            if not two:

                @nb.njit
                def fcn(x, X, Y):
                    return _interp_n_fg(ker0, ker1, x, X, Y)

            else:

                @nb.njit
                def fcn(x, X, Y, Z):
                    return _interp_n_fg_YZ(ker0, ker1, x, X, Y, Z)

    elif kind == "u":

        if lenker == 1:
            if not two:

                @nb.guvectorize(
                    [
                        (f4, f4[:], f4[:], f4[:]),
                        (f8, f8[:], f8[:], f8[:]),
                    ],
                    "(),(n),(n)->()",
                    nopython=True,
                )
                def fcn(x, X, Y, y):
                    y[0] = _interp_1(ker0, x, X, Y)

            else:

                @nb.guvectorize(
                    [
                        (f4, f4[:], f4[:], f4[:], f4[:], f4[:]),
                        (f8, f8[:], f8[:], f8[:], f8[:], f8[:]),
                    ],
                    "(),(n),(n),(n)->(),()",
                    nopython=True,
                )
                def fcn(x, X, Y, Z, y, z):
                    y[0], z[0] = _interp_1_YZ(ker0, x, X, Y, Z)

        elif lenker == 2:
            if not two:

                @nb.guvectorize(
                    [
                        (f4, f4[:], f4[:], f4[:], f4[:]),
                        (f8, f8[:], f8[:], f8[:], f8[:]),
                    ],
                    "(),(n),(n)->(),()",
                    nopython=True,
                )
                def fcn(x, X, Y, yf, yg):
                    yf[0], yg[0] = _interp_1_fg(ker0, ker1, x, X, Y)

            else:

                @nb.guvectorize(
                    [
                        (f4, f4[:], f4[:], f4[:], f4[:], f4[:], f4[:], f4[:]),
                        (f8, f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8[:]),
                    ],
                    "(),(n),(n),(n)->(),()",
                    nopython=True,
                )
                def fcn(x, X, Y, Z, yf, zf, yg, zg):
                    yf[0], zf[0], yg[0], zg[0] = _interp_1_fg_YZ(ker0, ker1, x, X, Y, Z)

    else:
        raise ValueError(f"Expected kind in ('u', '1', 'n'); got {kind}")

    return fcn


# Does not work.  Cannot index a tuple inside the guvectorize, it seems...
# def make_d(*f):
#     n = len(f)

#     @nb.guvectorize(
#         [(f8, f8[:], f8[:], nb.int64, f8[:])],
#         "(),(n),(n),()->()",
#     )
#     def fcn(x, X, Y, d, y):
#         if d < n:
#             y[0] = interp_1(x, X, Y, f[d])
#         else:
#             y[0] = 0.0

#     return fcn


def make_kernel(interpolant, deriv):
    """
    Select the interpolating kernel(s) for a given interpolation method and derivative.

    Parameters
    ----------
    interpolant : str, Default "linear"

        Returns interpolating kernel(s) for
            linear interpolation if "linear"
            Piecewise Cubic Hermite Interpolating Polynomial interpolation if "pchip"

    deriv : int or tuple of int, Default 0

        Return interpolating kernel for the `deriv` derivative of the `interpolant`.
        Note that `deriv = 0` simply builds a function that interpolates
        according to `interpolant`.
        If a tuple, returns a tuple of interpolating kernels, one for each
        of the derivatives in this tuple.

    """

    if interpolant == "linear":
        kers = (_linterp, _linterp1)
    elif interpolant == "pchip":
        kers = (_pchip, _pchip1, _pchip2, _pchip3)
    else:
        raise ValueError(
            f"Expected `interpolant` in ('linear', 'pchip'); got {interpolant}"
        )

    if isinstance(deriv, (tuple, list)):
        ker = tuple(kers[d] for d in deriv)
    else:
        ker = kers[deriv]

    return ker



"""Kernels for Piecewise Cubic Hermite Interpolating Polynomials (PCHIPs)"""
import numpy as np
import numba as nb


@nb.njit
def _pchip(x, X, Y, i):
    """
    The "kernel" of Piecewise Cubic Hermite Interpolating Polynomial (PCHIP) interpolation.

    Inputs and outputs analogous to `_linterp`.
    """

    s = x - X[i - 1]
    dY, cY, bY = _pchip_coeffs(X, Y, i)
    return Y[i - 1] + s * (dY + s * (cY + s * bY))


@nb.njit
def _pchip1(x, X, Y, i):
    """
    The "kernel" of the 1st derivative of PCHIP interpolation.

    Inputs and outputs analogous to `_linterp`.
    """
    s = x - X[i - 1]
    dY, cY, bY = _pchip_coeffs(X, Y, i)
    return dY + s * (2 * cY + 3 * s * bY)


@nb.njit
def _pchip2(x, X, Y, i):
    """
    The "kernel" of the 2nd derivative of PCHIP interpolation.

    Inputs and outputs analogous to `_linterp`.
    """
    s = x - X[i - 1]
    _, cY, bY = _pchip_coeffs(X, Y, i)
    return 2 * cY + 6 * s * bY


@nb.njit
def _pchip3(x, X, Y, i):
    """
    The "kernel" of the 3rd derivative of PCHIP interpolation.

    Inputs and outputs analogous to `_linterp`.
    """
    _, _, bY = _pchip_coeffs(X, Y, i)
    return 6 * bY


@nb.njit
def _pchipd(x, X, Y, i, d):

    if d == 0:
        return _pchip(x, X, Y, i)
    elif d == 1:
        return _pchip1(x, X, Y, i)
    elif d == 2:
        return _pchip2(x, X, Y, i)
    elif d == 3:
        return _pchip3(x, X, Y, i)
    else:
        return 0.0


@nb.njit
def _pchip_coeffs(X, Y, i):
    """
    Calculate the coefficients of a cubic interpolant

    Parameters
    ----------
    X, Y, i :
        As in `_pchip`

    Returns
    -------
    dY, cY, bY : float
        The first, second, and third order coefficients of the cubic interpolant,
        such that the value of the interpolant at `x` is
        `y = Y[i - 1] + s * (dY + s * (cY + s * bY))`

    """

    # Pre-assign sizes for PCHIP variables.
    h = [0.0, 0.0, 0.0]
    DY = [0.0, 0.0, 0.0]
    dY = [0.0, 0.0]

    # Check whether x is adjacent to the start or end of this X
    at_start = i == 1
    at_end = (i == len(X) - 1) or np.isnan(X[i + 1]) or np.isnan(Y[i + 1])

    if at_start and at_end:

        if np.isnan(X[i]) or np.isnan(Y[i]):
            # Only one valid data point.  Leave the interpolant as NaN.
            dY[0], cY, bY = np.nan, np.nan, np.nan

        else:
            # ||| X[0] <= x <= X[1] |||   Revert to Linear Interpolation

            dY[0] = (Y[i] - Y[i - 1]) / (X[i] - X[i - 1])
            # leave cY, bY = 0, 0

            # The following code evaluates the cubic interpolant, given `d` that
            # specifies the number of derivatives to take.
            # r = (x - X[i - 1]) / (X[i] - X[i - 1])
            # if d == 0:
            #     y = Y[i - 1] * (1 - r) + Y[i] * r
            # elif d == 1:
            #     y = (Y[i] - Y[i - 1]) / (X[i] - X[i - 1])
            # else:
            #     y = 0.0

    else:
        if at_start:
            #  ||| X[0] <= x <= X[1] < X[2] --->
            h[1] = X[i] - X[i - 1]
            h[2] = X[i + 1] - X[i]
            DY[1] = (Y[i] - Y[i - 1]) / h[1]
            DY[2] = (Y[i + 1] - Y[i]) / h[2]

            #  Noncentered, shape-preserving, three-point formula:
            dY[0] = ((2.0 * h[1] + h[2]) * DY[1] - h[1] * DY[2]) / (h[1] + h[2])
            if np.sign(dY[0]) != np.sign(DY[1]):
                dY[0] = 0.0
            elif (np.sign(DY[1]) != np.sign(DY[2])) and (
                np.abs(dY[0]) > np.abs(3.0 * DY[1])
            ):
                dY[0] = 3.0 * DY[1]

            # Standard PCHIP formula
            if np.sign(DY[1]) * np.sign(DY[2]) > 0.0:
                w1 = 2.0 * h[2] + h[1]
                w2 = h[2] + 2.0 * h[1]
                dY[1] = (w1 + w2) / (w1 / DY[1] + w2 / DY[2])
            else:
                dY[1] = 0.0

        elif at_end:
            # <--- X[i-2] < X[i-1] < x <= X[i] |||
            h[0] = X[i - 1] - X[i - 2]
            h[1] = X[i] - X[i - 1]
            DY[0] = (Y[i - 1] - Y[i - 2]) / h[0]
            DY[1] = (Y[i] - Y[i - 1]) / h[1]

            # Standard PCHIP formula
            if np.sign(DY[0]) * np.sign(DY[1]) > 0.0:
                w1 = 2.0 * h[1] + h[0]
                w2 = h[1] + 2.0 * h[0]
                dY[0] = (w1 + w2) / (w1 / DY[0] + w2 / DY[1])
            else:
                dY[0] = 0.0

            #  Noncentered, shape-preserving, three-point formula:
            dY[1] = ((h[0] + 2.0 * h[1]) * DY[1] - h[1] * DY[0]) / (h[0] + h[1])
            if np.sign(dY[1]) != np.sign(DY[1]):
                dY[1] = 0.0
            elif (np.sign(DY[1]) != np.sign(DY[0])) and (
                np.abs(dY[1]) > np.abs(3 * DY[1])
            ):

                dY[1] = 3.0 * DY[1]

        else:
            # <--- X[i-2] < X[i-1] < x <= X[i] < X[i+1] --->
            h[0] = X[i - 1] - X[i - 2]  # Way faster to do this
            h[1] = X[i] - X[i - 1]  # than
            h[2] = X[i + 1] - X[i]  # diff(X(i-2:i+1))
            DY[0] = (Y[i - 1] - Y[i - 2]) / h[0]
            DY[1] = (Y[i] - Y[i - 1]) / h[1]
            DY[2] = (Y[i + 1] - Y[i]) / h[2]

            # Standard PCHIP formula
            for j in range(2):
                if np.sign(DY[j]) * np.sign(DY[j + 1]) > 0.0:
                    w1 = 2.0 * h[j + 1] + h[j]
                    w2 = h[j + 1] + 2.0 * h[j]
                    dY[j] = (w1 + w2) / (w1 / DY[j] + w2 / DY[j + 1])
                else:
                    dY[j] = 0.0

        # Polynomial coefficients for this piece
        cY = (3.0 * DY[1] - 2.0 * dY[0] - dY[1]) / h[1]
        bY = (dY[0] - 2.0 * DY[1] + dY[1]) / h[1] ** 2

        # The following code evaluates the cubic interpolant, given `d` that
        # specifies the number of derivatives to take.
        # if d == 0:
        #     y = Y[i - 1] + s * (dY[0] + s * (cY + s * bY))
        # elif d == 1:  # first derivative
        #     y = dY[0] + s * (2 * cY + 3 * s * bY)
        # elif d == 2:  # second derivative
        #     y = 2 * cY + 6 * s * bY
        # elif d == 3:  # third derivative
        #     y = 6 * bY
        # else:
        #     y = 0.0
        # return y

    return dY[0], cY, bY



import numpy as np
import numba as nb

from .ppinterp import diff_1d_samesize


def pchip_coeffs(X, Y):
    """
    Piecewise Polynomial Coefficients for a PCHIP
    (Piecewise Cubic Hermite Interpolating Polynomial)

    Parameters
    ----------
    X : ndarray

        Independent variable. Must be same dimensions as `Y`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `Y`.

        `X` must be monotonically increasing in its last dimension.
        That is, `X[*i,:]` should be monotonically increasing for any
        `i` tuple indexing all but the last dimension.  NaN's in `X`
        are treated as +Inf.

    Y : ndarray

        Dependent variable.  Must be same dimensions as `X`, with any
        dimension possibly a singleton: that is, must be broadcastable
        to the size of `X`.

    Returns
    -------
    Yppc : ndarray

        Coefficients for a PCHIP interpolant of `Y` to `X`.

    Notes
    -----

    Evaluate the piecewise polynomial at `x` as

    >>> y = ppval(x, X, Yppc)

    """

    # Suppress the divide by zero and invalid arguments warnings that appear
    # for the `numba.guvectorize`d function, though not for the `numba.njit`ed
    # function.  The PCHIP algorithm is guaranteed to not divide by zero.
    # https://github.com/numba/numba/issues/4793#issuecomment-623296850
    with np.errstate(divide="ignore", invalid="ignore"):
        return _pchip_coeffs(X, Y, np.arange(4))


@nb.njit
def pchip_coeffs_1(X, Y):
    """
    Coefficients for a single PCHIP

    Inputs and outputs are as for `pchip_coeffs`, but `X` and `Y` both 1D
    arrays of length n, and so `Yppc` is a 2D array of size (n, 4).
    """

    # Note, np.diff does not work with nb.njit, here.
    # Also, using our own numba.njit'ed function is faster than np.ediff1d
    h = diff_1d_samesize(X)  # distance between data sites
    δ = diff_1d_samesize(Y) / h  # slope of linear interpolant between data sites

    K = X.size  # Number of data points included possible NaN's

    # Count the number of consecutive valid (non-NaN) data points.
    k = K
    for i in range(K):
        if np.isnan(X[i]) or np.isnan(Y[i]):
            k = i
            break

    d = np.zeros(k)  # slope of interpolant at data sites
    C = np.full((K, 4), np.nan)

    if k > 2:

        # Calculate PCHIP slopes
        #  Slopes at end points:
        #   Set d[0] and d[k-1] via non-centered, shape-preserving three-point formulae.
        #  Slopes at interior points:
        #   d[i] = weighted average of δ[i-1] and δ[i] when they have the same sign.
        #   d[i] = 0 when δ[i-1] and δ[i] have opposites signs or either is zero.

        d[0] = ((2 * h[0] + h[1]) * δ[0] - h[0] * δ[1]) / (h[0] + h[1])
        if np.sign(d[0]) != np.sign(δ[0]):
            d[0] = 0
        elif (np.sign(δ[0]) != np.sign(δ[1])) and (abs(d[0]) > abs(3 * δ[0])):
            d[0] = 3 * δ[0]

        for i in range(1, k - 1):
            if np.sign(δ[i - 1]) * np.sign(δ[i]) > 0:
                w1 = h[i - 1] + 2 * h[i]
                w2 = 2 * h[i - 1] + h[i]
                d[i] = (w1 + w2) / (w1 / δ[i - 1] + w2 / δ[i])
            else:
                d[i] = 0

        d[k - 1] = ((2 * h[k - 2] + h[k - 3]) * δ[k - 2] - h[k - 2] * δ[k - 3]) / (
            h[k - 2] + h[k - 3]
        )
        if np.sign(d[k - 1]) != np.sign(δ[k - 2]):
            d[k - 1] = 0
        elif (np.sign(δ[k - 2]) != np.sign(δ[k - 3])) and (
            abs(d[k - 1]) > abs(3 * δ[k - 2])
        ):
            d[k - 1] = 3 * δ[k - 2]

        # Build piecewise cubic Hermite polynomial
        for i in range(k - 1):
            dzzdx = (δ[i] - d[i]) / h[i]
            dzdxdx = (d[i + 1] - δ[i]) / h[i]
            C[i, 0] = (dzdxdx - dzzdx) / h[i]
            C[i, 1] = 2 * dzzdx - dzdxdx
            C[i, 2] = d[i]

    elif k == 2:
        # Special case: use linear interpolation.

        δ = (Y[1] - Y[0]) / (X[1] - X[0])

        C[0, 0] = 0
        C[0, 1] = 0
        C[0, 2] = δ

    # else:  # k == 1
    #     leave coefficients as nans.  Ignore case of x = X[0] while
    #     having X[1:] == np.nan and/or Y[1:] == np.nan.  That case
    #     cannot be handled here, since setting the coefficients to
    #     be [0, 0, 0, Y[0]], i.e. a constant function, would mean the
    #     interpolant would be Y[0] when evaluated at x > X[0], but we
    #     only want it to be Y[0] at x = X[0] precisely.

    C[:, 3] = Y

    return C


@nb.guvectorize([(nb.f8[:], nb.f8[:], nb.f8[:], nb.f8[:, :])], "(n),(n),(m)->(n,m)")
def _pchip_coeffs(X, Y, dummy, Yppc):
    Yppc[:, :] = pchip_coeffs_1(X, Y)



"""Neutral Tangent Plane and related functions"""

import numpy as np
import numba as nb

from neutralocean.lib import xr_to_np
from neutralocean.grid.graph import graph_binary_fcn
from neutralocean.eos.tools import make_eos_s_t


def ntp_ϵ_errors(s, t, p, grid, eos_s_t="gsw", grav=None, rho_c=None):
    """
    Calculate ϵ neutrality errors on an approximately neutral surface

    Parameters
    ----------
    s, t, p, eos_s_t, grav, rho_c :
        See `ntp_ϵ_errors_norms`.

    grid : dict
        See `ntp_ϵ_errors_norms`.
        Need not have the `distperp` element, as this is not used.
        If the `dist` element is missing, a value of 1.0 will be used.
        Can alternatively pass a 2 element tuple that is just `grid['edges']`,
        in which case `dist` will be taken as 1.0.


    Returns
    -------
    ϵ : array
        The ϵ neutrality errors on the surface.
        `ϵ[i]` is the neutrality error between nodes `a[i]` and `b[i]`, where
        `a, b = grid['edges']`.
    """

    if isinstance(grid, tuple):
        edges = grid
        dist = 1.0
    elif isinstance(grid, dict):
        edges = grid["edges"]
        dist = grid.get("dist", 1.0)

    eos_s_t = make_eos_s_t(eos_s_t, grav, rho_c)

    s, t, p = (xr_to_np(x) for x in (s, t, p))

    sa, ta, pa = (graph_binary_fcn(edges, x, avg1) for x in (s, t, p))
    ds, dt = (graph_binary_fcn(edges, x, dif1) for x in (s, t))

    rsa, rta = eos_s_t(sa, ta, pa)
    ϵ = rsa * ds + rta * dt
    if dist is not float or dist != 1.0:
        ϵ = ϵ / dist
    return ϵ


def ntp_ϵ_errors_norms(s, t, p, grid, eos_s_t="gsw", grav=None, rho_c=None):
    """
    Calculate norms of the ϵ neutrality errors on an approximately neutral surface

    Parameters
    ----------
    s, t, p : ndarray
        practical / Absolute Salinity, potential / Conservative temperature,
        and pressure / depth on the surface

    grid : dict
        Containing the following:

        edges : tuple of length 2, Required
            Each element is an array of int of length E, where E is the number of
            edges in the grid's graph, i.e. the number of pairs of adjacent water
            columns (including land) in the grid.
            If `edges = (a, b)`, the nodes (water columns) whose linear indices are
            `a[i]` and `b[i]` are adjacent.
            If one of `a[i]` or `b[i]` is less than 0, that edge does not exist. # TODO:  Is this necessary?

        dist : 1d array, Default 1.0
            Horizontal distance between adjacent water columns (nodes).
            `dist[i]` is the distance between nodes whose linear indices are
            `edges[0][i]` and `edges[1][i]`.

        distperp : 1d array, Default 1.0
            Horizontal distance of the face between adjacent water columns (nodes).
            `distperp[i]` is the distance of the interface between nodes whose
            linear indices are `edges[0][i]` and `edges[1][i]`.

    eos_s_t : function
        Equation of state for the partial derivatives of density or specific
        volume with respect to `S` and `T` as a function of `S`, `T`, and `P`
        inputs.

    grav : float, Default None
        Gravitational acceleration [m s-2].  When non-Boussinesq, pass None.

    rho_c : float, Default None
        Boussinesq reference desnity [kg m-3].  When non-Boussinesq, pass None.

    Returns
    -------
    ϵ_RMS : float
        Area-weighted root-mean-square of the ϵ neutrality error on the surface

    ϵ_MAV : TYPE
        Area-weighted mean-absolute-value of the ϵ neutrality error on the surface

    """

    eos_s_t = make_eos_s_t(eos_s_t, grav, rho_c)

    # Calculate ϵ neutrality errors.  Here, treat all distances = 1.
    # The actual distances will be handled in computing the norms
    ϵ = ntp_ϵ_errors(s, t, p, {"edges": grid["edges"]}, eos_s_t)

    area = grid["dist"] * grid["distperp"]  # Area [m^2] centred on edges

    # L2 norm of vector [a_i], weighted by vector [w_i], is sqrt( sum( w_i * a_i^2 ) / sum( w_i ) )
    # Here, weights are `area`.
    # But also need to divide epsilon by grid distances `dist`.
    # Thus, the numerator of L2 norm needs to multiply epsilon^2 by
    #     area / dist^2 = distperp / dist,
    ϵ_RMS = np.sqrt(
        np.nansum(grid["distperp"] / grid["dist"] * ϵ**2)
        / np.sum(area * np.isfinite(ϵ))
    )

    # L1 norm of vector [a_i], weighted by vector [w_i], is sum( w_i * |a_i| ) / sum( w_i )
    # Here, weights are `area`.
    # But also need to divide epsilon by grid distances `dist`.
    # Thus, the numerator of L1 norm needs to multiply epsilon by
    #     area ./ dist = distperp ,
    ϵ_MAV = np.nansum(grid["distperp"] * abs(ϵ)) / np.sum(
        area * np.isfinite(ϵ)
    )

    return ϵ_RMS, ϵ_MAV


@nb.njit
def avg1(a, b):
    return (a + b) * 0.5


@nb.njit
def dif1(a, b):
    return a - b



# In[Imports]

# Functions to make the Equation of State
from neutralocean.eos import make_eos, make_eos_s_t

# Functions to compute various approximately neutral surfaces
from neutralocean.surface import potential_surf, anomaly_surf, omega_surf

# Functions to load OCCA data
from neutralocean.examples.load_OCCA import load_OCCA

from neutralocean.grid.rectilinear import build_grid, edgedata_to_maps

# In[Load OCCA data]

# Load OCCA data from the local folder containing this script.
# If it is not there, it will be downloaded to that folder.
# If that folder does not have write permissions, it will be downloaded to temp
# files, loaded into memory, then the temp files deleted.
g, S, T = load_OCCA()  # S, T arranged as (Longitude, Latitude, Depth)
ni, nj, nk = S.shape
Z = -g["RC"]  # Depth vector (note positive and increasing down)

# Select pinning cast in the middle of the domain
i0 = int(ni / 2)
j0 = int(nj / 2)
z0 = 1500.0

# make Boussinesq version of the Jackett and McDougall (1995) equation of state
# --- which is what OCCA used --- and its partial derivatives
eos = make_eos("jmd95", g["grav"], g["ρ_c"])
eos_s_t = make_eos_s_t("jmd95", g["grav"], g["ρ_c"])

# Build grid adjacency and distance information for neutralocean functions
grid = build_grid(
    (ni, nj), g["wrap"], g["DXCvec"], g["DYCsc"], g["DYGsc"], g["DXGvec"]
)

# In[Potential Density surfaces]

# Provide reference pressure (actually depth, in Boussinesq) and isovalue
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos="jmd95",
    vert_dim="Depth_c",
    ref=0.0,
    isoval=1027.5,
)
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" with isovalue = {d['isoval']}kg m-3"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS']} kg m-4"
)

# Provide pin_cast and pin_p: the reference location and depth that the surface intersects
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos="jmd95",
    vert_dim="Depth_c",
    ref=0.0,
    pin_cast=(i0, j0),
    pin_p=z0,
)
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS']} kg m-4"
)

# Provide just the location to intersect `(pin_cast, pin_p)`.
# This takes the reference depth `ref` to match `pin_p`.
# Also illustrate using xarray coordinates for pin_cast.
# Also use PCHIPs as the vertical interpolants.
# Also provide the equation of state and its partial derivatives w.r.t S and T.
s, t, z, d = potential_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    pin_cast={"Longitude_t": 180.5, "Latitude_t": 0.5},
    pin_p=z0,
    interp="pchip",
)
print(
    f" ** The potential density surface (referenced to {d['ref']}m)"
    f" intersecting the cast at (180.5 E, 0.5 N) at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS']} kg m-4"
)

# In[Delta surfaces]

# Provide reference salinity and potential temperature values
s0, t0 = 34.5, 4.0
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    ref=(s0, t0),
    isoval=0.0,
)
print(
    f" ** The in-situ density anomaly surface (referenced to {d['ref']})"
    f" with isovalue = {d['isoval']}kg m-3"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS']} kg m-4"
)

# Provide pin_cast and pin_p: the reference location and depth that the surface intersects
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    ref=(s0, t0),
    pin_cast=(i0, j0),
    pin_p=z0,
)
print(
    f" ** The in-situ density anomaly surface (referenced to {d['ref']})"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS']} kg m-4"
)

# Provide just the location to intersect: depth `pin_p` on cast `pin_cast`
# This takes the reference S and T values from that location.
s, t, z, d = anomaly_surf(
    S,
    T,
    Z,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
)
print(
    f" ** The in-situ density anomaly surface (referenced to {d['ref']})"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" (isovalue = {d['isoval']}kg m-3)"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS']} kg m-4"
)

# In[Omega surfaces]

# Initialize omega surface with a (locally referenced) potential density surface.
# Provide grid distances.
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
    eos=(eos, eos_s_t),
    ITER_MAX=10,
    ITER_START_WETTING=1,
)
print(
    f" ** The omega-surface"
    f" initialized from a potential density surface (referenced to {z0}m)"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS'][-1]} kg m-4"
)

# Initialize omega surface with a (locally referenced) in-situ density anomaly surface.
# Use PCHIP interpolation rather than the default, linear interpolation.
# Remove the mixed layer, calculated internally according to the given parameters --
#   see `mixed_layer` for details on these parameters.
s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    ref=(None, None),
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
    eos=(eos, eos_s_t),
    interp="pchip",
    p_ml={"bottle_index": 1, "ref_p": 0.0},
    ITER_MAX=10,
    ITER_START_WETTING=1,
    TOL_P_SOLVER=1e-5,
)
print(
    f" ** The omega-surface"
    f" initialized from an in-situ density anomaly surface (referenced locally to cast {(i0,j0)} at {z0}m)"
    f" intersecting the cast indexed by {(i0,j0)} at depth {z0}m"
    f" has root-mean-square ϵ neutrality error {d['ϵ_RMS'][-1]} kg m-4"
)


# In[Begin showing more advanced features]

import numpy as np
from neutralocean.mixed_layer import mixed_layer
from neutralocean.ntp import ntp_ϵ_errors, ntp_ϵ_errors_norms
from neutralocean.label import veronis_density
from neutralocean.lib import _process_casts, find_first_nan
from neutralocean.interp1d import make_interpolator
from neutralocean.ppinterp import select_ppc
from neutralocean.eos import make_eos_p, vectorize_eos
from neutralocean.traj import ntp_bottle_to_cast, _ntp_bottle_to_cast

# In[Veronis Density, used to label an approx neutral surface]
S_ref_cast = S.values[i0, j0]
T_ref_cast = T.values[i0, j0]
ρ_v = veronis_density(S_ref_cast, T_ref_cast, Z, z0, eos="jmd95")
print(
    f"A surface through the cast indexed by {(i0,j0)} at depth {z0}m"
    f" has Veronis density {ρ_v} kg m-3"
)

# In[Remove mixed layer from an omega surface]

# Pre-compute depth of the mixed layer
z_ml = mixed_layer(S, T, Z, eos)

s, t, z, d = omega_surf(
    S,
    T,
    Z,
    grid,
    ref=(None, None),
    vert_dim="Depth_c",
    pin_cast=(i0, j0),
    pin_p=z0,
    eos=(eos, eos_s_t),
    ITER_MAX=10,
    ITER_START_WETTING=1,
    TOL_P_SOLVER=1e-5,
    p_ml=z_ml,
)

# potential and anomaly surfaces don't take `p_ml` as an argument, since they
# aren't iterative algorithms, so we can just remove it manually, e.g.
# z[z < z_ml] = np.nan

# In[Neutrality errors on a surface]
ϵ_RMS, ϵ_MAV = ntp_ϵ_errors_norms(s, t, z, grid, eos_s_t)
print(f"RMS of ϵ is {ϵ_RMS : 4e} [kg m-4])")

# Calculate ϵ neutrality errors on all pairs of adjacent water columns
ϵ = ntp_ϵ_errors(s, t, z, grid, eos_s_t)

# Convert ϵ above into two 2D maps, one for zonal ϵ errors and one for meridional ϵ errors
ϵx, ϵy = edgedata_to_maps(ϵ, (ni, nj), (True, False))
# These can then be mapped...

# In[Neutral Tangent Plane bottle to cast]

sB, tB, zB = 35.0, 16.0, 500.0  # Thermodynamic properties of a given Bottle
S1 = S.values[180, 80, :]
T1 = T.values[180, 80, :]
s1, t1, z1 = ntp_bottle_to_cast(sB, tB, zB, S1, T1, Z)

# Or the more manual version:
n_good = find_first_nan(S1)[()]
ppc_fn = select_ppc("linear", "1")
S1ppc, T1ppc = (ppc_fn(Z, C) for C in (S1, T1))
s1, t1, z1 = _ntp_bottle_to_cast(
    sB, tB, zB, S1ppc, T1ppc, Z, n_good, 1e-4, eos
)


# In[Work with Numpy arrays instead of xarrays]

# Convert S and T from xarray to numpy ndarrays, and make vertical dimension
# contiguous in memory.  If not done here in advance, this will be done each
# time potential_surf, anomaly_surf, or omega_surf is called.  Hence, this is
# the recommended method if many approx neutral surfaces will be calculated.
Snp, Tnp, Znp = _process_casts(S, T, Z, "Depth_c")

s, t, z, d = anomaly_surf(
    Snp,
    Tnp,
    Znp,
    grid=grid,
    eos=(eos, eos_s_t),
    vert_dim=-1,
    ref=(s0, t0),
    isoval=0.0,
)


# In[Calculate a large-scale potential vorticity on our surface]

# Create function for partial deriv of equation of state with respect to depth z
eos_z = make_eos_p("jmd95", g["grav"], g["ρ_c"])  # for scalar inputs
eos_z = vectorize_eos(eos_z)  # for nd inputs

# Earth sidereal day period [s]
Earth_day = 86164

# Coriolis param [s-1] on tracer grid
f = 2 * (2 * np.pi / Earth_day) * np.sin(g["YCvec"] * (np.pi / 180))

linterp_dx_utwo = make_interpolator("linear", deriv=1, kind="u", two=True)
sz, tz = linterp_dx_utwo(
    z, Z, S.values, T.values
)  # ∂S/∂Z and ∂T/∂Z, on the surface
rs, rt = eos_s_t(s, t, z)  # ∂ρ/∂S and ∂ρ/∂T, on the surface

# ∂δ/∂z on the surface, where δ is the in-situ density anomaly
δz = rs * sz + rt * tz + (eos_z(s, t, z) - eos_z(s0, t0, z))

# large-scale potential vorticity [m-1 s-1] defined via in-situ density anomaly, on the surface
q = f * δz
